                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*version 5: functional continuations*/
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipDictionary.h"
#include "SlipEvaluate.h"
#include "SlipThread.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  ((sizeof(TYP) - sizeof(CNT))/ sizeof(CEL_type))

/*------------------------------ private constants ---------------------------*/

static const TXT_type AL1_error_string = "at least one expression required";
static const TXT_type BRQ_error_string = "boolean value required";
static const TXT_type E1X_error_string = "exactly one expression required";
static const TXT_type ILT_error_string = "illegal expression type";
static const TXT_type ITA_error_string = "improperly terminated application";
static const TXT_type ITF_error_string = "improperly terminated form";
static const TXT_type ITP_error_string = "improperly terminated procedure";
static const TXT_type IPA_error_string = "invalid parameter";
static const TXT_type IVP_error_string = "invalid pattern in form";
static const TXT_type IVV_error_string = "missing variable in form";
static const TXT_type IXT_error_string = "invalid expression type";
static const TXT_type MSP_error_string = "missing pattern in form";
static const TXT_type MSV_error_string = "invalid variable in form";
static const TXT_type PNR_error_string = "procedure or native required";
static const TXT_type TFO_error_string = "too few operands in application";
static const TXT_type TFX_error_string = "too few expressions in form";
static const TXT_type TMO_error_string = "too many operands in application";
static const TXT_type TMX_error_string = "too many expressions in form";

/*------------------------------ private prototypes --------------------------*/

static EXP_type        evaluate_bindings(PRC_type,
                                         PAI_type);
static EXP_type            evaluate_body(PRC_type,
                                         ENV_type);
static EXP_type      evaluate_expression(EXP_type);
static EXP_type     evaluate_native_call(NAT_type,
                                         PAI_type);
static EXP_type        evaluate_sequence(PAI_type);
static EXP_type evaluate_vararg_bindings(PRC_type,
                                         SYM_type,
                                         PAI_type,
                                         ENV_type);

/*------------------------------ private functions ---------------------------*/

static EXP_type continue_with(EXP_type Value)
  { CNT_type continuation;
    CCC_type function;
    CFN_type c_function;
    continuation = Thread_Pop();
    c_function = continuation->cfn;
    function = c_function->ccc;
    return function(Value,
                    continuation); }

static SGN_type list_length(PAI_type List)
  { PAI_type pair;
    SGN_type count;
    count = 0;
    for (pair = List;
         is_PAI(pair);
         pair = pair->cdr)
      count++;
    if (is_NUL(pair))
      return count;
    return -1; }

/*----------------------------------------------------------------------------*/
/*--------------------------------- application ------------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Application_String = "application";

static CFN_type Continue_application;

typedef struct aPL * aPL_type;
typedef struct aPL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type opd; } aPL;

static const UNS_type aPL_size = chunk_size(aPL);

static EXP_type continue_application(EXP_type Procedure,
                                     CNT_type Continuation)
  { aPL_type application_thread;
    PAI_type operands;
    TAG_type tag;
    application_thread = (aPL_type)Continuation;
    operands = application_thread->opd;
    tag = Tag_of(Procedure);
    switch (tag)
      { case NAT_tag:
          return evaluate_native_call(Procedure,
                                      operands); 
        case PRC_tag:
             return evaluate_bindings(Procedure,
                                      operands); }
    return Main_Error_Text(PNR_error_string,
                           Application_String); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_application(EXP_type Operator,
                                     EXP_type Operands)
  { aPL_type application_thread;
    CNT_type continuation;
    continuation = Thread_Push(Continue_application,
                               aPL_size);
    application_thread = (aPL_type)continuation;
    application_thread->opd = Operands;
    return evaluate_expression(Operator); }

static NIL_type initialize_application(NIL_type)
  { Continue_application = make_CFN(continue_application); }
    
/*----------------------------------------------------------------------------*/
/*------------------------------------ begin ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_begin(PAI_type Operands)
  { TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(AL1_error_string,
                                 Main_Begin_String); 
        case PAI_tag:
          return evaluate_sequence(Operands); }
    return Main_Error_Text(ITF_error_string,
                           Main_Begin_String); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- bindings --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_bindings;

typedef struct bND * bND_type;
typedef struct bND { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     PAI_type par;
                     PAI_type opd;
                     ENV_type clo; } bND;

static const UNS_type bND_size = chunk_size(bND);

static EXP_type binding(PRC_type Procedure,
                        EXP_type Parameters,
                        PAI_type Operands,
                        ENV_type Closure)
  { bND_type binding_thread;
    CNT_type continuation;
    ENV_type closure;
    EXP_type expression;
    PAI_type operands;
    TAG_type tag;
    tag = Tag_of(Parameters);
    switch (tag)
      { case NUL_tag:
          if (is_NUL(Operands))
            return evaluate_body(Procedure,
                                 Closure); 
          return Main_Error_Procedure(TMO_error_string,
                                      Procedure);
        case PAI_tag:
          tag = Tag_of(Operands);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Procedure(TFO_error_string,
                                            Procedure);
              case PAI_tag:
                expression = Operands->car;
                operands   = Operands->cdr;
                continuation = Thread_Push(Continue_bindings,
                                           bND_size);
                binding_thread = (bND_type)continuation;
                binding_thread->prc = Procedure;
                binding_thread->par = Parameters;
                binding_thread->opd = operands;
                binding_thread->clo = Closure;
                return evaluate_expression(expression); }
          return Main_Error_Procedure(ITA_error_string,
                                      Procedure);
        case SYM_tag:
          tag = Tag_of(Operands);
          switch (tag)
            { case NUL_tag:
                closure = Dictionary_Bind(Parameters,
                                          Main_Empty_Environment,
                                          Closure);
                return evaluate_body(Procedure,
                                     closure);
              case PAI_tag:
                return evaluate_vararg_bindings(Procedure,
                                                Parameters,
                                                Operands,
                                                Closure); }
          return Main_Error_Procedure(ITA_error_string,
                                      Procedure); }
    return Main_Error_Procedure(ITP_error_string,
                                Procedure); }

static EXP_type continue_bindings(EXP_type Value,
                                  CNT_type Continuation)
  { bND_type binding_thread;
    ENV_type closure;
    EXP_type parameter;
    PAI_type operands,
             parameters;
    PRC_type procedure;
    binding_thread = (bND_type)Continuation;
    procedure    = binding_thread->prc;
    parameters   = binding_thread->par;
    operands     = binding_thread->opd;
    closure      = binding_thread->clo;
    parameter = parameters->car;
    if (is_SYM(parameter))
      { parameters = parameters->cdr;
        closure = Dictionary_Bind(parameter,
                                  Value,
                                  closure);
        return binding(procedure,
                       parameters,
                       operands,
                       closure); }
    return Main_Error_Procedure(IPA_error_string,
                                procedure); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_bindings(PRC_type Procedure,
                                  PAI_type Operands)
  { ENV_type closure;
    EXP_type parameters;
    parameters = Procedure->par;
    closure    = Procedure->clo;
    return binding(Procedure,
                   parameters,
                   Operands,
                   closure); }
                
static NIL_type initialize_bindings(NIL_type)
  { Continue_bindings = make_CFN(continue_bindings); }
    
/*----------------------------------------------------------------------------*/
/*-------------------------------------- body --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_body;

typedef struct bOD * bOD_type;
typedef struct bOD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     ENV_type env; } bOD;

static const UNS_type bOD_size = chunk_size(bOD);

static EXP_type continue_body(EXP_type Value,
                              CNT_type Continuation)
  { bOD_type body_thread;
    body_thread = (bOD_type)Continuation;
    Dictionary_Environment = body_thread->env;
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_body(PRC_type Procedure,
                              ENV_type Closure)
  { bOD_type body_thread;
    CNT_type continuation;
    PAI_type body;
    continuation = Thread_Push(Continue_body,
                               bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Dictionary_Environment;
    Dictionary_Environment = Closure;
    body = Procedure->bod;
    return evaluate_sequence(body); }

static EXP_type evaluate_inline(EXP_type Expression)
  { bOD_type body_thread;
    CNT_type continuation;
    continuation = Thread_Push(Continue_body,
                               bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Dictionary_Environment;
    return evaluate_expression(Expression); }

static EXP_type evaluate_inline_sequence(PAI_type Expressions)
  { bOD_type body_thread;
    CNT_type continuation;
    continuation = Thread_Push(Continue_body,
                               bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Dictionary_Environment;
    return evaluate_sequence(Expressions); }

static NIL_type initialize_body(NIL_type)
  { Continue_body = make_CFN(continue_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ define --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_define;

typedef struct dEF * dEF_type;
typedef struct dEF { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } dEF;

static const UNS_type dEF_size = chunk_size(dEF);

static EXP_type continue_define(EXP_type Value,
                                CNT_type Continuation)
  { dEF_type define_thread;
    SYM_type variable;
    define_thread = (dEF_type)Continuation;
    variable = define_thread->var;
    Dictionary_Define(variable,
                      Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_define(PAI_type Operands)
  { dEF_type define_thread;
    CNT_type continuation;
    EXP_type expression,
             pattern;
    PAI_type body,
             expressions,
             parameters,
             residue;
    PRC_type procedure;
    SYM_type variable;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(MSP_error_string,
                                 Main_Define_String);
        case PAI_tag:
          pattern = Operands->car;
          body    = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(AL1_error_string,
                                       Main_Define_String);
              case PAI_tag:
                tag = Tag_of(pattern);
                switch (tag)
                  { case PAI_tag:
                      expressions = (PAI_type)pattern;
                      variable = expressions->car;
                      if (is_SYM(variable))
                        { parameters = expressions->cdr;
                          procedure = make_PRC(variable,
                                               parameters,
                                               body,
                                               Main_Empty_Environment);
                          Dictionary_Define(variable,
                                            procedure);
                          procedure->clo = Dictionary_Environment;
                          return procedure; }
                      return Main_Error_Text(IVV_error_string,
                                             Main_Define_String);
                    case SYM_tag:
                      variable = (SYM_type)pattern;
                      expression = body->car;
                      residue    = body->cdr;
                      if (is_NUL(residue))
                        { continuation = Thread_Push(Continue_define,
                                                     dEF_size);
                          define_thread = (dEF_type)continuation;
                          define_thread->var = variable;
                          return evaluate_expression(expression); }
                      return Main_Error_Text(E1X_error_string,
                                             Main_Define_String); }
                return Main_Error_Text(IVP_error_string,
                                       Main_Define_String); }
          return Main_Error_Text(ITF_error_string,
                                 Main_Define_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Define_String); }

static NIL_type initialize_define(NIL_type)
  { Continue_define = make_CFN(continue_define); }

/*----------------------------------------------------------------------------*/
/*------------------------------- direct_application -------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Direct_Application_String = "direct application";

static EXP_type evaluate_direct_application(EXP_type Value,
                                            PAI_type Arguments)
  { EXP_type pattern,
             value;
    ENV_type closure;
    FUN_type function;
    NAT_type native;
    PAI_type arguments,
             parameters;
    PRC_type procedure;
    SGN_type length;
    SYM_type variable;
    TAG_type tag;
    UNS_type index;
    VEC_type vector;
    tag = Tag_of(Value);
    switch (tag)
      { case NAT_tag:
          native = (NAT_type)Value;
          function = native->fun;
          length = list_length(Arguments);
          switch (length)
            { case 0:
                vector = Main_Empty_Vector;
                break;
              case -1:
                { TXT_type name = native->nam;
                  return Main_Error_Text(ITA_error_string,
                                         name); }
              default:
                { arguments = Arguments;
                  vector = make_VEC(length);
                  for (index = 1;
                       index <= length;
                       index++)
                    { value = arguments->car;
                      arguments = arguments->cdr;
                      vector[index] = value; }
                  break; }}
          return function(vector); 
        case PRC_tag:
          arguments = Arguments;
          procedure = (PRC_type)Value;
          pattern = procedure->par;
          closure = procedure->clo;
          for (;;)
            { tag = Tag_of(pattern);
              switch (tag)
                { case NUL_tag:
                    if (is_NUL(arguments))
                      return evaluate_body(procedure,
                                           closure);
                    return Main_Error_Procedure(TMO_error_string,
                                                procedure);
                  case PAI_tag:
                    tag = Tag_of(arguments);
                    switch (tag)
                      { case NUL_tag:
                          return Main_Error_Procedure(TFO_error_string,
                                                      procedure);
                        case PAI_tag:
                          value = arguments->car;
                          parameters = pattern;
                          variable = parameters->car;
                          if (is_SYM(variable))
                            { arguments = arguments->cdr;
                              pattern   = parameters->cdr;
                              closure = Dictionary_Bind(variable,
                                                        value,
                                                        closure);
                              continue; }
                          return Main_Error_Procedure(IPA_error_string,
                                                      procedure); }
                    return Main_Error_Procedure(ITA_error_string,
                                                procedure);
                  case SYM_tag:
                    variable = pattern;
                    closure = Dictionary_Bind(variable,
                                              arguments,
                                              closure);
                    return evaluate_body(procedure,
                                         closure); }
              return Main_Error_Procedure(ITP_error_string,
                                          procedure); }}
    return Main_Error_Text(PNR_error_string,
                           Direct_Application_String); }
                    
/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_double;

typedef struct iFD * iFD_type;
typedef struct iFD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns;
                     EXP_type alt; } iFD;

static const UNS_type iFD_size = chunk_size(iFD);

static EXP_type continue_if_double(EXP_type Boolean,
                                   CNT_type Continuation)
  { iFD_type if_double_thread;
    EXP_type alternative,
             consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_double_thread = (iFD_type)Continuation;
    switch (tag)
      { case FLS_tag:
          alternative = if_double_thread->alt;
          return evaluate_inline(alternative);
        case TRU_tag:
          consequent = if_double_thread->cns;
          return evaluate_inline(consequent); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }

/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_single;

typedef struct iFS * iFS_type;
typedef struct iFS { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns; } iFS;

static const UNS_type iFS_size = chunk_size(iFS);

static EXP_type continue_if_single(EXP_type Boolean,
                                   CNT_type Continuation)
  { iFS_type if_single_thread;
    EXP_type consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_single_thread = (iFS_type)Continuation;
    switch (tag)
      { case FLS_tag:
          return Main_Unspecified;
        case TRU_tag:
          consequent = if_single_thread->cns;
          return evaluate_inline(consequent); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_if(PAI_type Operands)
  { iFD_type if_double_thread;
    iFS_type if_single_thread;
    CNT_type continuation;
    EXP_type alternative,
             consequent,
             predicate,
             residue;
    PAI_type alternatives,
             clauses;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_If_String);
        case PAI_tag:
          predicate = Operands->car;
          clauses   = Operands->cdr;
          tag = Tag_of(clauses);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(TFX_error_string,
                                       Main_If_String);
              case PAI_tag:
                consequent   = clauses->car;
                alternatives = clauses->cdr;
                tag = Tag_of(alternatives);
                switch (tag)
                  { case NUL_tag:
                      continuation = Thread_Push(Continue_if_single,
                                                 iFS_size);
                      if_single_thread = (iFS_type)continuation;
                      if_single_thread->cns = consequent;
                      return evaluate_expression(predicate);
                    case PAI_tag:
                      alternative = alternatives->car;
                      residue     = alternatives->cdr;
                      if (is_NUL(residue))
                        { continuation = Thread_Push(Continue_if_double,
                                                     iFD_size);
                          if_double_thread = (iFD_type)continuation;
                          if_double_thread->cns = consequent;
                          if_double_thread->alt = alternative;
                          return evaluate_expression(predicate); }
                      return Main_Error_Text(TMX_error_string,
                                             Main_If_String); }
                return Main_Error_Text(ITF_error_string,
                                       Main_If_String); }
          return Main_Error_Text(ITF_error_string,
                                 Main_If_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_If_String); }

static NIL_type initialize_if(NIL_type)
  { Continue_if_double = make_CFN(continue_if_double);
    Continue_if_single = make_CFN(continue_if_single); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_lambda(PAI_type Operands)
  { EXP_type parameters;
    PAI_type body;
    PRC_type procedure;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_Lambda_String);
        case PAI_tag:
          parameters = Operands->car;
          body       = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(AL1_error_string,
                                       Main_Lambda_String);
              case PAI_tag:
                procedure = make_PRC(Main_Anonymous,
                                     parameters,
                                     body,
                                     Dictionary_Environment);
                return procedure; }
          return Main_Error_Text(ITF_error_string,
                                 Main_Lambda_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Lambda_String); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- native call -------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_native_call;

typedef struct nCL * nCL_type;
typedef struct nCL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type exs;
                     VEC_type vec;
                     NAT_type nat;
                     NBR_type pos; } nCL;

static const UNS_type nCL_size = chunk_size(nCL);

static EXP_type native_call(PAI_type Expressions,
                            VEC_type Vector,
                            NAT_type Native,
                            UNS_type Index)
  { nCL_type native_call_thread;
    CNT_type continuation;
    EXP_type expression;
    NBR_type position;
    PAI_type expressions;
    expression  = Expressions->car;
    expressions = Expressions->cdr;
    position = make_NBR(Index);
    continuation = Thread_Push(Continue_native_call,
                               nCL_size);
    native_call_thread = (nCL_type)continuation;
    native_call_thread->exs = expressions;
    native_call_thread->vec = Vector;
    native_call_thread->nat = Native;
    native_call_thread->pos = position;
    return evaluate_expression(expression); }

static EXP_type continue_native_call(EXP_type Value,
                                     CNT_type Continuation)
  { nCL_type native_call_thread;
    FUN_type function;
    NAT_type native;
    NBR_type position;
    PAI_type expressions;
    UNS_type index,
             length;
    VEC_type vector;
    native_call_thread = (nCL_type)Continuation;
    vector       = native_call_thread->vec;
    native       = native_call_thread->nat;
    position     = native_call_thread->pos;
    index = position->lng,
    length = size_VEC(vector);
    vector[index] = Value;
    if (index < length)
      { expressions = native_call_thread->exs;
        return native_call(expressions,
                           vector,
                           native,
                           index + 1); }
    function = native->fun;
    return function(vector); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_native_call(NAT_type Native,
                                     PAI_type Operands)
  { FUN_type function;
    SGN_type length;
    TXT_type name;
    VEC_type vector;
    length = list_length(Operands);
    switch (length)
      { case 0:
          function = Native->fun;
          return function(Main_Empty_Vector);
        case -1:
          name = Native->nam;
          return Main_Error_Text(ITA_error_string,
                                 name); }
    vector = make_VEC(length);
    return native_call(Operands,
                       vector,
                       Native,
                       1); }

static NIL_type initialize_native_call(NIL_type)
  { Continue_native_call = make_CFN(continue_native_call); }
  
/*----------------------------------------------------------------------------*/
/*------------------------------------ quote ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_quote(PAI_type Operands)
  { EXP_type expression,
             residue;
    expression = Operands->car;
    residue    = Operands->cdr;
    if (is_NUL(residue))
      return expression;            /* clone !!! */
    return Main_Error_Text(TMX_error_string,
                           Main_Quote_String); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_sequence;

typedef struct sEQ * sEQ_type;
typedef struct sEQ { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type exs; } sEQ;

static const UNS_type sEQ_size = chunk_size(sEQ);

static EXP_type continue_sequence(EXP_type Value,
                                  CNT_type Continuation)
  { sEQ_type sequence_thread;
    PAI_type expressions;
    sequence_thread = (sEQ_type)Continuation;
    expressions = sequence_thread->exs;
    return evaluate_sequence(expressions); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_sequence(PAI_type Expressions)
  { sEQ_type sequence_thread;
    CNT_type continuation;
    EXP_type expression;
    PAI_type expressions;
    TAG_type tag;
    expression  = Expressions->car;
    expressions = Expressions->cdr;
    tag = Tag_of(expressions);
    switch (tag)
      { case NUL_tag:
          return evaluate_expression(expression);
        case PAI_tag:
          continuation = Thread_Push(Continue_sequence,
                                     sEQ_size);
          sequence_thread = (sEQ_type)continuation;
          sequence_thread->exs = expressions;
          return evaluate_expression(expression); }
    return Main_Error_Text(ITF_error_string,
                           Main_Sequence_String); }

static NIL_type initialize_sequence(NIL_type)
  { Continue_sequence = make_CFN(continue_sequence); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_set;

typedef struct sET * sET_type;
typedef struct sET { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } sET;

static const UNS_type sET_size = chunk_size(sET);

static EXP_type continue_set(EXP_type Value,
                             CNT_type Continuation)
  { sET_type set_thread;
    SYM_type variable;
    set_thread = (sET_type)Continuation;
    variable = set_thread->var;
    Dictionary_Replace(variable,
                       Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set(PAI_type Operands)
  { sET_type set_thread;
    CNT_type continuation;
    EXP_type expression;
    PAI_type expressions,
             residue;
    SYM_type variable;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(MSV_error_string,
                                 Main_Set_String);
        case PAI_tag:
          variable = Operands->car;
          if (is_SYM(variable))
            { expressions = Operands->cdr;
              tag = Tag_of(expressions);
              switch (tag)
                { case NUL_tag:
                    return Main_Error_Text(E1X_error_string,
                                           Main_Set_String);
                  case PAI_tag:
                    expression = expressions->car;
                    residue    = expressions->cdr;
                    if (is_NUL(residue))
                      { continuation = Thread_Push(Continue_set,
                                                   sET_size);
                        set_thread = (sET_type)continuation;
                        set_thread->var = variable;
                        return evaluate_expression(expression); }
                    return Main_Error_Text(TMX_error_string,
                                           Main_Set_String); }
              return Main_Error_Text(ITF_error_string,
                                     Main_Set_String); }
          return Main_Error_Text(IVV_error_string,
                                 Main_Set_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_Set_String); }

static NIL_type initialize_set(NIL_type)
  { Continue_set = make_CFN(continue_set); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ symbol --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_symbol(SYM_type Variable)
  { EXP_type value;
    value = Dictionary_Lookup(Variable);
    return value; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- value ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_value(EXP_type Value)
  { return Value; }

/*----------------------------------------------------------------------------*/
/*------------------------------- vararg_bindings ----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_vararg_bindings;

typedef struct vRG * vRG_type;
typedef struct vRG { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     SYM_type par;
                     PAI_type opd;
                     PAI_type arg;
                     ENV_type clo; } vRG;

static const UNS_type vRG_size = chunk_size(vRG);

static EXP_type vararg_bindings(PRC_type Procedure,
                                SYM_type Parameter,
                                PAI_type Operands,
                                EXP_type Arguments,
                                ENV_type Closure)
  { vRG_type vararg_bindings_thread;
    CNT_type continuation;
    ENV_type closure;
    EXP_type expression;
    PAI_type arguments,
             operands;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          arguments = Main_Reverse(Arguments);
          closure = Dictionary_Bind(Parameter,
                                    arguments,
                                    Closure);
          return evaluate_body(Procedure,
                               closure);
        case PAI_tag:
          expression = Operands->car;
          operands   = Operands->cdr;
          continuation = Thread_Push(Continue_vararg_bindings,
                                     vRG_size);
          vararg_bindings_thread = (vRG_type)continuation;
          vararg_bindings_thread->prc = Procedure;
          vararg_bindings_thread->par = Parameter;
          vararg_bindings_thread->opd = operands;
          vararg_bindings_thread->arg = Arguments;
          vararg_bindings_thread->clo = Closure;
          return evaluate_expression(expression); }
    return Main_Error_Procedure(ITA_error_string,
                                Procedure); }

static EXP_type continue_vararg_bindings(EXP_type Value,
                                         CNT_type Continuation)
  { vRG_type vararg_bindings_thread;
    ENV_type closure;
    PAI_type arguments,
             operands;
    PRC_type procedure;
    SYM_type parameter;
    vararg_bindings_thread = (vRG_type)Continuation;
    procedure    = vararg_bindings_thread->prc;
    parameter    = vararg_bindings_thread->par;
    operands     = vararg_bindings_thread->opd;
    arguments    = vararg_bindings_thread->arg;
    closure      = vararg_bindings_thread->clo;
    arguments = make_PAI(Value,
                         arguments);
    return vararg_bindings(procedure,
                           parameter,
                           operands,
                           arguments,
                           closure); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_vararg_bindings(PRC_type Procedure,
                                         SYM_type Parameter,
                                         PAI_type Operands,
                                         ENV_type Closure)
  { return vararg_bindings(Procedure,
                           Parameter,
                           Operands,
                           Main_Empty_Pair,
                           Closure); }

static NIL_type initialize_vararg_bindings(NIL_type)
  { Continue_vararg_bindings = make_CFN(continue_vararg_bindings); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_while_body;
static CFN_type Continue_while_predicate;

typedef struct wHB * wHB_type;
typedef struct wHB { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     PAI_type bod;
                     ENV_type env; } wHB;

static const UNS_type wHB_size = chunk_size(wHB);

typedef struct wHP * wHP_type;
typedef struct wHP { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     EXP_type bod;
                     ENV_type env;
                     EXP_type res; } wHP;

static const UNS_type wHP_size = chunk_size(wHP);

static EXP_type while_predicate(EXP_type Predicate,
                                PAI_type Body,
                                ENV_type Environment,
                                EXP_type Result)
  { wHP_type while_predicate_thread;
    CNT_type continuation;
    continuation = Thread_Push(Continue_while_predicate,
                               wHP_size);
    while_predicate_thread = (wHP_type)continuation;
    while_predicate_thread->prd = Predicate;
    while_predicate_thread->bod = Body;
    while_predicate_thread->env = Environment;
    while_predicate_thread->res = Result;
    return evaluate_inline(Predicate); }

static EXP_type continue_while_body(EXP_type Result,
                                    CNT_type Continuation)
  { wHB_type while_body_thread;
    ENV_type environment;
    EXP_type predicate;
    PAI_type body;
    while_body_thread = (wHB_type)Continuation;
    predicate   = while_body_thread->prd;
    body        = while_body_thread->bod;
    environment = while_body_thread->env;
    Dictionary_Environment = environment;
    return while_predicate(predicate,
                           body,
                           Dictionary_Environment,
                           Result); }

static EXP_type continue_while_predicate(EXP_type Boolean,
                                    CNT_type Continuation)
  { wHB_type while_body_thread;
    wHP_type while_predicate_thread;
    CNT_type continuation;
    ENV_type environment;
    EXP_type predicate,
             result;
    PAI_type body;
    while_predicate_thread = (wHP_type)Continuation;
    predicate    = while_predicate_thread->prd;
    body         = while_predicate_thread->bod;
    environment  = while_predicate_thread->env;
    Dictionary_Environment = environment;
    if (is_FLS(Boolean))
      { result = while_predicate_thread->res;
        return result; }
    continuation = Thread_Push(Continue_while_body,
                               wHB_size);
    while_body_thread = (wHB_type)continuation;
    while_body_thread->prd = predicate;
    while_body_thread->bod = body;
    while_body_thread->env = Dictionary_Environment;
    return evaluate_inline_sequence(body); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_while(PAI_type Operands)
  { EXP_type body,
             predicate;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          return Main_Error_Text(TFX_error_string,
                                 Main_While_String);
        case PAI_tag:
          predicate = Operands->car;
          body   = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                return Main_Error_Text(TFX_error_string,
                                       Main_While_String);
              case PAI_tag:
                return while_predicate(predicate,
                                       body,
                                       Dictionary_Environment,
                                       Main_Unspecified); }
          return Main_Error_Text(ITF_error_string,
                                 Main_While_String); }
    return Main_Error_Text(ITF_error_string,
                           Main_While_String); }

static NIL_type initialize_while(NIL_type)
  { Continue_while_predicate = make_CFN(continue_while_predicate);
    Continue_while_body = make_CFN(continue_while_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ form ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_form(PAI_type Form)
  { EXP_type operands,
             operator;
    operator = Form->car;
    operands = Form->cdr;
    if (operator == Main_Begin)
          return evaluate_begin(operands);
    if (operator == Main_Define)
         return evaluate_define(operands);
    if (operator == Main_If)
             return evaluate_if(operands);
    if (operator == Main_Lambda)
         return evaluate_lambda(operands);
    if (operator == Main_Quote)
          return evaluate_quote(operands);
    if (operator == Main_Set)
            return evaluate_set(operands);
    if (operator == Main_While)
          return evaluate_while(operands);
    return evaluate_application(operator,
                                operands); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_expression(EXP_type Expression)
  { TAG_type tag;
    tag = Tag_of(Expression);
    switch (tag)
      { case PAI_tag:
            return evaluate_form(Expression);
        case SYM_tag:
          return evaluate_symbol(Expression);
        case CHA_tag:
        case FLS_tag:
        case NUL_tag:
        case NBR_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
           return evaluate_value(Expression); 
        case CFN_tag:
        case CNT_tag:
        case ENV_tag:
        case NAT_tag:
        case PRC_tag:
        case USP_tag:
          Main_Error_Tag(ILT_error_string,
                         tag); }
    return Main_Error_Tag(IXT_error_string,
                          tag); }

/*------------------------------ public functions ----------------------------*/

NIL_type Evaluate_Initialize(NIL_type)
  { initialize_application();
    initialize_bindings();
    initialize_body();
    initialize_define();
    initialize_if();
    initialize_native_call();
    initialize_sequence();
    initialize_set();
    initialize_vararg_bindings();
    initialize_while(); }

EXP_type Evaluate_Continue(EXP_type Value)
  { return continue_with(Value); }

EXP_type Evaluate_Apply(EXP_type Procedure,
                        PAI_type Arguments)
  { return evaluate_direct_application(Procedure,
                                       Arguments); }

EXP_type Evaluate_Evaluate(EXP_type Expression)
  { return evaluate_expression(Expression); }
